@@ -443,17 +443,24 @@ get_label (KeySym keyval)
 
         }
     }
-        return label;
+    return label;
 }
 
 static PyObject *
-virtkey_get_labels_from_keycode_internal(virtkey *cvirt, long keycode)
+virtkey_get_labels_from_keycode_internal(virtkey *cvirt, long keycode,
+                                         const long* mod_masks, int num_masks)
 {
-    static const int mods[] = {0,1,2,128,129};
+    static const long mods[] = {0,1,2,128,129};
     int m;
     unsigned int mods_rtn;
     KeySym keysym = 0;
-    PyObject * labelTuple = PyTuple_New(5);
+
+    if (mod_masks == NULL)
+    {
+        mod_masks = mods;
+        num_masks = (sizeof(mods)/sizeof(*mods));
+    }
+    PyObject* labelTuple = PyTuple_New(num_masks);
 
     /* get effective group index */
     XkbStateRec state;
@@ -462,10 +469,10 @@ virtkey_get_labels_from_keycode_internal(virtkey *cvirt, long keycode)
         XkbIsLegalGroup(state.locked_group))  /* be defensive */
         group = state.locked_group;
 
-    for(m = 4; m >= 0; --m)
+    for(m = 0; m < num_masks; m++)
     {
         if (XkbTranslateKeyCode (cvirt->kbd, (KeyCode) keycode,
-                                 XkbBuildCoreState (mods[m], group),
+                                 XkbBuildCoreState (mod_masks[m], group),
                                  &mods_rtn, &keysym)){
             PyTuple_SetItem(labelTuple,m,get_label(keysym));
         }
@@ -475,6 +482,40 @@ virtkey_get_labels_from_keycode_internal(virtkey *cvirt, long keycode)
     return labelTuple;
 }
 
+static PyObject *
+virtkey_get_keysyms_from_keycode_internal(virtkey *cvirt, long keycode,
+                                         const long* mod_masks, int num_masks)
+{
+    static const long mods[] = {0};
+    int m;
+    unsigned int mods_rtn;
+    KeySym keysym = 0;
+
+    if (mod_masks == NULL)
+    {
+        mod_masks = mods;
+        num_masks = (sizeof(mods)/sizeof(*mods));
+    }
+    PyObject* tuple = PyTuple_New(num_masks);
+
+    /* get effective group index */
+    XkbStateRec state;
+    int group = 0;
+    if (Success == XkbGetState(cvirt->display, XkbUseCoreKbd, &state) &&
+        XkbIsLegalGroup(state.locked_group))  /* be defensive */
+        group = state.locked_group;
+
+    for(m = 0; m < num_masks; m++)
+    {
+        if (!XkbTranslateKeyCode (cvirt->kbd, (KeyCode) keycode,
+                                  XkbBuildCoreState (mod_masks[m], group),
+                                  &mods_rtn, &keysym)){
+            keysym = 0;
+        }
+        PyTuple_SetItem(tuple, m, PyLong_FromLong(keysym));
+    }
+    return tuple;
+}
 
 static PyObject *
 report_key_info (virtkey * cvirt, XkbKeyPtr key, int col, int *x, int *y)
@@ -518,7 +559,8 @@ report_key_info (virtkey * cvirt, XkbKeyPtr key, int col, int *x, int *y)
         if (!strncmp (name, cvirt->kbd->names->keys[k].name, XkbKeyNameLength))
         {
 
-            labelTuple = virtkey_get_labels_from_keycode_internal(cvirt, k);
+            labelTuple = virtkey_get_labels_from_keycode_internal(cvirt, k,
+                                                                  NULL, 0);
 
 
             x1 = *x + geom->shapes[key->shape_ndx].bounds.x1/10;
@@ -722,11 +764,95 @@ virtkey_get_labels_from_keycode(PyObject *self, PyObject *args)
 {
     virtkey *cvirt = (virtkey *)self;
     long keycode;
+    PyObject* omod_masks = NULL;
+    long* mod_masks = NULL;
+    int num_masks = 0;
+    int i;
+    PyObject* results = NULL;
 
-    if (!(PyArg_ParseTuple(args, "i", &keycode)))
+    if (!(PyArg_ParseTuple(args, "i|O", &keycode, &omod_masks)))
         return NULL;
 
-    return virtkey_get_labels_from_keycode_internal(cvirt, keycode);
+    if (omod_masks)
+    {
+        if (!PySequence_Check(omod_masks))
+        {
+            PyErr_SetString(PyExc_ValueError, "expected sequence type");
+        }
+        else
+        {
+            num_masks = PySequence_Length(omod_masks);
+            mod_masks = PyMem_Malloc(num_masks * sizeof(*mod_masks));
+            for (i = 0; i < num_masks; i++)
+            {
+                PyObject* item = PySequence_GetItem(omod_masks, i);
+                if (!PyLong_Check(item))
+                {
+                    PyErr_SetString(PyExc_ValueError, "expected integer number");
+                    break;
+                }
+                mod_masks[i] = PyLong_AsLong(item);
+            }
+        }
+
+    }
+
+    if (!PyErr_Occurred())
+        results = virtkey_get_labels_from_keycode_internal(cvirt, keycode,
+                                                           mod_masks, num_masks);
+    if (mod_masks)
+        PyMem_Free(mod_masks);
+
+    return results;
+
+}
+
+static PyObject *
+virtkey_get_keysyms_from_keycode(PyObject *self, PyObject *args)
+{
+    virtkey *cvirt = (virtkey *)self;
+    long keycode;
+    PyObject* omod_masks = NULL;
+    long* mod_masks = NULL;
+    int num_masks = 0;
+    int i;
+    PyObject* results = NULL;
+
+    if (!(PyArg_ParseTuple(args, "i|O", &keycode, &omod_masks)))
+        return NULL;
+
+    if (omod_masks)
+    {
+        if (!PySequence_Check(omod_masks))
+        {
+            PyErr_SetString(PyExc_ValueError, "expected sequence type");
+        }
+        else
+        {
+            num_masks = PySequence_Length(omod_masks);
+            mod_masks = PyMem_Malloc(num_masks * sizeof(*mod_masks));
+            for (i = 0; i < num_masks; i++)
+            {
+                PyObject* item = PySequence_GetItem(omod_masks, i);
+                if (!PyLong_Check(item))
+                {
+                    PyErr_SetString(PyExc_ValueError, "expected integer number");
+                    break;
+                }
+                mod_masks[i] = PyLong_AsLong(item);
+            }
+        }
+
+    }
+
+    if (!PyErr_Occurred())
+        results = virtkey_get_keysyms_from_keycode_internal(cvirt, keycode,
+                                                          mod_masks, num_masks);
+    if (mod_masks)
+        PyMem_Free(mod_masks);
+
+    return results;
+
 }
 
 static PyObject * virtkey_send_unicode(PyObject * self,PyObject *args, Bool press){
@@ -945,6 +1071,64 @@ static PyObject * virtkey_get_current_group_name(PyObject * self,
     return result;
 }
 
+static PyObject * virtkey_get_current_group(PyObject * self,
+                                            PyObject * noargs)
+{
+    PyObject * result = NULL;
+    virtkey * cvirt  = (virtkey *)self;
+    Display * display = cvirt->display;
+
+    /* get current group index */
+    XkbStateRec state;
+    if (Success != XkbGetState(display, XkbUseCoreKbd, &state))
+        PyErr_SetString(virtkey_error, "XkbGetState failed");
+    else{
+        int group = state.locked_group;
+        if (!XkbIsLegalGroup(group))  /* be defensive */
+            PyErr_SetString(virtkey_error,
+                            "invalid effective group");
+        else
+        {
+            result = PyInt_FromLong(group);
+        }
+    }
+
+    if (PyErr_Occurred())
+        return NULL;
+
+    if (!result)
+        Py_RETURN_NONE;
+
+    return result;
+}
+
+// Reads the contents of the root window property _XKB_RULES_NAMES.
+static PyObject * virtkey_get_rules_names(PyObject *self, PyObject *args)
+{
+    PyObject* result = NULL;
+    virtkey * cvirt  = (virtkey *)self;
+    XkbRF_VarDefsRec vd;
+    char *tmp = NULL;
+
+    Display * display = cvirt->display;
+
+    if (XkbRF_GetNamesProp(display, &tmp, &vd) && tmp)
+    {
+        result = PyTuple_New(5);
+        PyTuple_SetItem(result, 0, PyUnicode_FromString(tmp ? tmp : ""));
+        PyTuple_SetItem(result, 1, PyUnicode_FromString(vd.model ? vd.model : ""));
+        PyTuple_SetItem(result, 2, PyUnicode_FromString(vd.layout ? vd.layout : ""));
+        PyTuple_SetItem(result, 3, PyUnicode_FromString(vd.variant ? vd.variant : ""));
+        PyTuple_SetItem(result, 4, PyUnicode_FromString(vd.options ? vd.options : ""));
+    }
+    else
+    {
+        result = PyTuple_New(0);
+    }
+
+    return result;
+}
+
 /* returns a plus-sign separated string of all keyboard layouts */
 static PyObject * virtkey_get_layout_symbols(PyObject * self,
                                              PyObject * noargs)
@@ -990,8 +1174,11 @@ static PyMethodDef virtkey_methods[] = {
   {"layout_get_section_size", virtkey_layout_get_section_info, METH_VARARGS},
   {"get_layouts", virtkey_get_layouts, METH_VARARGS},
   {"labels_from_keycode", virtkey_get_labels_from_keycode, METH_VARARGS},
+  {"keysyms_from_keycode", virtkey_get_keysyms_from_keycode, METH_VARARGS},
   {"reload", virtkey_reload_kbd, METH_VARARGS},
   {"get_current_group_name", virtkey_get_current_group_name, METH_NOARGS},
+  {"get_current_group", virtkey_get_current_group, METH_NOARGS},
+  {"get_rules_names", virtkey_get_rules_names, METH_NOARGS},
   {"get_layout_symbols", virtkey_get_layout_symbols, METH_NOARGS},
   {NULL, NULL},
 };
