@@ -35,6 +35,7 @@
 #include <X11/Xlib.h>
 #include <X11/extensions/XTest.h>
 #include <X11/XKBlib.h>
+#include <X11/extensions/XKBrules.h>
 #include <X11/extensions/XKBgeom.h>
 
 #define N_MODIFIER_INDEXES (Mod5MapIndex + 1)
@@ -74,9 +75,12 @@ static PyObject * virtkey_press_keycode(PyObject * self,PyObject *args);
 static PyObject * virtkey_release_keycode(PyObject * self,PyObject *args);
 
 static PyObject *
-virtkey_get_labels_from_keycode_internal(virtkey *cvirt, long keycode);
+virtkey_get_labels_from_keycode_internal(virtkey *cvirt, long keycode,
+                                         const long* mod_masks, int num_masks);
 static PyObject *
 virtkey_get_labels_from_keycode(PyObject *self, PyObject *args);
+static PyObject *
+virtkey_get_keysyms_from_keycode(PyObject *self, PyObject *args);
 
 static PyObject * virtkey_press_keysym(PyObject * self,PyObject *args);
 static PyObject * virtkey_release_keysym(PyObject * self,PyObject *args);
